import logging
from typing import Optional

import click

from copernicusmarine.catalogue_parser.models import CopernicusMarineCatalogue
from copernicusmarine.command_line_interface.exception_handler import (
    log_exception_and_exit,
)
from copernicusmarine.command_line_interface.utils import tqdm_disable_option
from copernicusmarine.core_functions import documentation_utils
from copernicusmarine.core_functions.click_custom_class import (
    CustomClickOptionsCommand,
)
from copernicusmarine.core_functions.describe import describe_function
from copernicusmarine.core_functions.fields_query_builder import (
    build_query,
    get_queryable_requested_fields,
)

logger = logging.getLogger("copernicusmarine")
blank_logger = logging.getLogger("copernicusmarine_blank_logger")


@click.group()
def cli_describe() -> None:
    pass


@cli_describe.command(
    "describe",
    cls=CustomClickOptionsCommand,
    short_help="Print products metadata of Copernicus Marine catalogue as JSON.",
    help=documentation_utils.DESCRIBE["DESCRIBE_DESCRIPTION_HELP"]
    + " \n\nReturns\n "
    + documentation_utils.DESCRIBE["DESCRIBE_RESPONSE_HELP"],  # noqa
    epilog="""
    Examples:

    .. code-block:: bash

        copernicusmarine describe --contains METOFFICE-GLO-SST-L4-NRT-OBS-SST-V2 --return-fields datasets

    .. code-block:: bash

        copernicusmarine describe -c METOFFICE-GLO-SST-L4-NRT-OBS-SST-V2 \n
    """,  # noqa
)
@click.option(
    "--show-all-versions",
    type=bool,
    is_flag=True,
    default=False,
    help=documentation_utils.DESCRIBE["SHOW_ALL_VERSIONS_HELP"],
)
@click.option(
    "--return-fields",
    "-r",
    type=str,
    default=None,
    help=documentation_utils.DESCRIBE["RETURN_FIELDS_HELP"],
)
@click.option(
    "--exclude-fields",
    "-e",
    type=str,
    default=None,
    help=documentation_utils.DESCRIBE["EXCLUDE_FIELDS_HELP"],
)
@click.option(
    "--contains",
    "-c",
    type=str,
    multiple=True,
    help=documentation_utils.DESCRIBE["CONTAINS_HELP"],
)
@click.option(
    "--product-id",
    "-p",
    type=str,
    default=None,
    help=documentation_utils.DESCRIBE["PRODUCT_ID_HELP"],
)
@click.option(
    "--dataset-id",
    "-i",
    type=str,
    default=None,
    help=documentation_utils.DESCRIBE["DATASET_ID_HELP"],
)
@click.option(
    "--max-concurrent-requests",
    type=int,
    default=15,
    help=documentation_utils.DESCRIBE["MAX_CONCURRENT_REQUESTS_HELP"],
)
@tqdm_disable_option
@click.option(
    "--log-level",
    type=click.Choice(["DEBUG", "INFO", "WARN", "ERROR", "CRITICAL", "QUIET"]),
    default="INFO",
    help=documentation_utils.DESCRIBE["LOG_LEVEL_HELP"],
)
@click.option(
    "--staging",
    type=bool,
    default=False,
    is_flag=True,
    hidden=True,
)
@log_exception_and_exit
def describe(
    show_all_versions: bool,
    return_fields: Optional[str],
    exclude_fields: Optional[str],
    contains: list[str],
    product_id: Optional[str],
    dataset_id: Optional[str],
    max_concurrent_requests: int,
    disable_progress_bar: bool,
    log_level: str,
    staging: bool,
) -> None:
    if log_level == "QUIET":
        logger.disabled = True
        logger.setLevel(level="CRITICAL")
    else:
        logger.setLevel(level=log_level)

    if logger.isEnabledFor(logging.DEBUG):
        logger.debug("DEBUG mode activated")

    response_catalogue = describe_function(
        show_all_versions,
        contains,
        product_id,
        dataset_id,
        max_concurrent_requests,
        disable_progress_bar,
        staging,
    )
    include_query, exclude_query = _create_include_and_exclude(
        return_fields,
        exclude_fields,
    )
    blank_logger.info(
        response_catalogue.model_dump_json(
            exclude_unset=True,
            exclude_none=True,
            exclude=exclude_query,
            include=include_query,
            indent=2,
            context={"sort_keys": False},
        )
    )


def _create_include_and_exclude(
    return_fields: Optional[str],
    exclude_fields: Optional[str],
) -> tuple[Optional[dict], Optional[dict]]:
    include_query = None
    exclude_query = None
    if return_fields:
        include_in_output = set(return_fields.replace(" ", "").split(","))
        if "all" not in include_in_output:
            queryable_fields = get_queryable_requested_fields(
                include_in_output,
                CopernicusMarineCatalogue,
                "--return-fields",
            )
            include_query = build_query(
                queryable_fields, CopernicusMarineCatalogue
            )
    if exclude_fields:
        exclude_from_output = set(exclude_fields.replace(" ", "").split(","))
        queryable_fields = get_queryable_requested_fields(
            exclude_from_output, CopernicusMarineCatalogue, "--exclude-fields"
        )
        exclude_query = build_query(
            queryable_fields, CopernicusMarineCatalogue
        )

    return include_query, exclude_query



import base64
import configparser
import logging
import pathlib
from netrc import netrc
from platform import system
from typing import Literal, Optional, Union

import click
import lxml.html
import requests

from copernicusmarine.core_functions.environment_variables import (
    COPERNICUSMARINE_CREDENTIALS_DIRECTORY,
    COPERNICUSMARINE_SERVICE_PASSWORD,
    COPERNICUSMARINE_SERVICE_USERNAME,
)
from copernicusmarine.core_functions.sessions import (
    BearerAuth,
    get_configured_requests_session,
)

logger = logging.getLogger("copernicusmarine")

USER_DEFINED_CACHE_DIRECTORY: str = (
    COPERNICUSMARINE_CREDENTIALS_DIRECTORY or ""
)
DEFAULT_CLIENT_BASE_DIRECTORY: pathlib.Path = (
    pathlib.Path(USER_DEFINED_CACHE_DIRECTORY)
    if USER_DEFINED_CACHE_DIRECTORY
    else pathlib.Path.home()
) / ".copernicusmarine"
DEFAULT_CLIENT_CREDENTIALS_FILENAME = ".copernicusmarine-credentials"
DEFAULT_CLIENT_CREDENTIALS_FILEPATH = (
    DEFAULT_CLIENT_BASE_DIRECTORY / DEFAULT_CLIENT_CREDENTIALS_FILENAME
)
RECOVER_YOUR_CREDENTIALS_MESSAGE = (
    "Learn how to recover your credentials at: "
    "https://help.marine.copernicus.eu/en/articles/"
    "4444552-i-forgot-my-username-or-my-password-what-should-i-do"
)

COPERNICUS_MARINE_AUTH_SYSTEM_DOMAIN = "auth.marine.copernicus.eu"
COPERNICUS_MARINE_AUTH_SYSTEM_URL = (
    f"https://{COPERNICUS_MARINE_AUTH_SYSTEM_DOMAIN}/"
)
COPERNICUS_MARINE_AUTH_SYSTEM_TOKEN_ENDPOINT = (
    COPERNICUS_MARINE_AUTH_SYSTEM_URL
    + "realms/MIS/protocol/openid-connect/token"
)
COPERNICUS_MARINE_AUTH_SYSTEM_USERINFO_ENDPOINT = (
    COPERNICUS_MARINE_AUTH_SYSTEM_URL
    + "realms/MIS/protocol/openid-connect/userinfo"
)

COPERNICUS_MARINE_MARINE_AUTH_OLD_SYSTEM_DOMAIN = "cmems-cas.cls.fr"

COPERNICUS_MARINE_MARINE_AUTH_OLD_SYSTEM_URL = (
    f"https://{COPERNICUS_MARINE_MARINE_AUTH_OLD_SYSTEM_DOMAIN}/cas/login"
)
ACCEPTED_HOSTS_NETRC_FILE = [
    "nrt.cmems-du.eu",
    "my.cmems-du.eu",
    COPERNICUS_MARINE_AUTH_SYSTEM_DOMAIN,
    "default_host",
]
DEPRECATED_HOSTS = [
    "nrt.cmems-du.eu",
    "my.cmems-du.eu",
]


class CredentialsCannotBeNone(Exception):
    """
    Exception raised when credentials are not set.

    To use the Copernicus Marine Service, you need to provide a username and
    a password. You can set them as environment variables or pass them as
    arguments to the function or use the :func:`~copernicusmarine.login` command.
    To register and create your valid credentials, please visit the
    Copernicus Marine `registration page <https://data.marine.copernicus.eu/register>`_
    """

    pass


class InvalidUsernameOrPassword(Exception):
    """
    Exception raised when the username or password are invalid.

    To register and create your valid credentials, please visit:
    Copernicus Marine `registration page <https://data.marine.copernicus.eu/register>`_
    """

    pass


class CouldNotConnectToAuthenticationSystem(Exception):
    """
    Exception raised when the client could not connect to the authentication system.

    Please check the following common problems:

    - Check your internet connection
    - Make sure to authorize ``cmems-cas.cls.fr`` and/or ``auth.marine.copernicus.eu`` domains

    If none of this worked, maybe the authentication system is down, please try again later.
    """  # noqa

    pass


def _warning_netrc_deprecated_hosts():
    logger.warning(
        "The following hosts are deprecated and will be removed in future versions: "
        f"{DEPRECATED_HOSTS}. "
        "Please update your netrc file to use the new authentication system domain: "
        f"{COPERNICUS_MARINE_AUTH_SYSTEM_DOMAIN}."
    )


def _warning_motuclient_deprecated():
    logger.warning(
        "The motuclient configuration file is deprecated "
        "and will be removed in future versions. Please use "
        "the login command and or function to create a new configuration file."
    )


def _load_credential_from_copernicus_marine_configuration_file(
    credential_type: Literal["username", "password"],
    configuration_filename: pathlib.Path,
) -> Optional[str]:
    configuration_file = open(configuration_filename)
    configuration_string = base64.standard_b64decode(
        configuration_file.read()
    ).decode("utf8")
    config = configparser.RawConfigParser()
    config.read_string(configuration_string)
    credential = config.get("credentials", credential_type)
    if credential:
        logger.debug(f"{credential_type} loaded from {configuration_filename}")
    return credential


def _load_credential_from_netrc_configuration_file(
    credential_type: Literal["username", "password"],
    configuration_filename: pathlib.Path,
) -> Optional[str]:
    authenticator = None
    for host in ACCEPTED_HOSTS_NETRC_FILE:
        authenticator = netrc(configuration_filename).authenticators(host=host)
        if (
            authenticator
            and host in DEPRECATED_HOSTS
            and credential_type == "username"
        ):
            _warning_netrc_deprecated_hosts()
        if authenticator:
            break
    if authenticator:
        username, _, password = authenticator
        logger.debug(f"{credential_type} loaded from {configuration_filename}")
        return username if credential_type == "username" else password
    else:
        return None


def _load_credential_from_motu_configuration_file(
    credential_type: Literal["username", "password"],
    configuration_filename: pathlib.Path,
) -> Optional[str]:
    motu_file = open(configuration_filename)
    motu_credential_type = "user" if credential_type == "username" else "pwd"
    config = configparser.RawConfigParser()
    config.read_string(motu_file.read())
    credential = config.get("Main", motu_credential_type)
    if credential:
        logger.debug(f"{credential_type} loaded from {configuration_filename}")
    return credential


def _retrieve_credential_from_prompt(
    credential_type: Literal["username", "password"], hide_input: bool
) -> str:
    if credential_type == "username":
        logger.info(
            "Downloading Copernicus Marine data requires a Copernicus Marine username "
            "and password, sign up for free at:"
            " https://data.marine.copernicus.eu/register"
        )
    return click.prompt(
        "Copernicus Marine " + credential_type, hide_input=hide_input
    )


def _retrieve_credential_from_environment_variable(
    credential_type: Literal["username", "password"],
) -> Optional[str]:
    if credential_type == "username":
        logger.debug("Tried to load username from environment variable")
        return COPERNICUSMARINE_SERVICE_USERNAME
    if credential_type == "password":
        logger.debug("Tried to load password from environment variable")
        return COPERNICUSMARINE_SERVICE_PASSWORD


def _retrieve_credential_from_custom_configuration_files(
    credential_type: Literal["username", "password"],
    credentials_file: pathlib.Path,
) -> Optional[str]:
    if "netrc" in str(credentials_file):
        credential = _load_credential_from_netrc_configuration_file(
            credential_type, credentials_file
        )
    elif "motuclient" in str(credentials_file):
        if credential_type == "username":
            _warning_motuclient_deprecated()
        credential = _load_credential_from_motu_configuration_file(
            credential_type, credentials_file
        )
    else:
        credential = (
            _load_credential_from_copernicus_marine_configuration_file(
                credential_type, credentials_file
            )
        )
    return credential


def _retrieve_credential_from_default_configuration_files(
    credential_type: Literal["username", "password"],
) -> Optional[str]:
    copernicus_marine_configuration_file = pathlib.Path(
        DEFAULT_CLIENT_CREDENTIALS_FILEPATH
    )
    motu_configuration_file = pathlib.Path(
        pathlib.Path.home() / "motuclient" / "motuclient-python.ini"
    )
    netrc_configuration_file = pathlib.Path(
        pathlib.Path.home() / ("_netrc" if system() == "Windows" else ".netrc")
    )
    if copernicus_marine_configuration_file.exists():
        credential = (
            _load_credential_from_copernicus_marine_configuration_file(
                credential_type,
                copernicus_marine_configuration_file,
            )
        )
    elif motu_configuration_file.exists():
        if credential_type == "username":
            _warning_motuclient_deprecated()
        credential = _load_credential_from_motu_configuration_file(
            credential_type, motu_configuration_file
        )
    elif netrc_configuration_file.exists():
        credential = _load_credential_from_netrc_configuration_file(
            credential_type, netrc_configuration_file
        )
    else:
        credential = None
    return credential


def _retrieve_credential_from_configuration_files(
    credential_type: Literal["username", "password"],
    credentials_file: Optional[pathlib.Path],
) -> Optional[str]:
    if credentials_file and credentials_file.exists():
        credential = _retrieve_credential_from_custom_configuration_files(
            credential_type, credentials_file
        )
    else:
        credential = _retrieve_credential_from_default_configuration_files(
            credential_type
        )
    return credential


def copernicusmarine_configuration_file_exists(
    configuration_file_directory: pathlib.Path,
) -> bool:
    configuration_filename = pathlib.Path(
        configuration_file_directory / DEFAULT_CLIENT_CREDENTIALS_FILENAME
    )
    return configuration_filename.exists()


def copernicusmarine_validate_and_get_user(
    configuration_file: Optional[pathlib.Path],
    username: Optional[str],
    password: Optional[str],
) -> Union[str, None]:
    if username and password:
        if user := _validate_and_get_user(username, password):
            logger.info("Valid credentials from input username and password.")
            return user
        else:
            logger.error(
                "Invalid credentials from input username and password."
            )
            logger.info(RECOVER_YOUR_CREDENTIALS_MESSAGE)
            return None
    elif (
        COPERNICUSMARINE_SERVICE_USERNAME and COPERNICUSMARINE_SERVICE_PASSWORD
    ):
        if user := _validate_and_get_user(
            COPERNICUSMARINE_SERVICE_USERNAME,
            COPERNICUSMARINE_SERVICE_PASSWORD,
        ):
            logger.info(
                "Valid credentials from environment variables: "
                "COPERNICUSMARINE_SERVICE_USERNAME and "
                "COPERNICUSMARINE_SERVICE_PASSWORD."
            )
            return user
        else:
            logger.error(
                "Invalid credentials from environment variables: "
                "COPERNICUSMARINE_SERVICE_USERNAME and "
                "COPERNICUSMARINE_SERVICE_PASSWORD."
            )
            logger.info(RECOVER_YOUR_CREDENTIALS_MESSAGE)
            return None
    elif (
        username := _retrieve_credential_from_configuration_files(
            "username", configuration_file
        )
    ) and (
        password := _retrieve_credential_from_configuration_files(
            "password", configuration_file
        )
    ):
        if user := _validate_and_get_user(username, password):
            logger.info("Valid credentials from configuration file.")
            return user
        else:
            logger.error("Invalid credentials from configuration file.")
            logger.info(RECOVER_YOUR_CREDENTIALS_MESSAGE)
            return None
    elif configuration_file:
        logger.info(
            f"No credentials found in configuration file {configuration_file}."
        )
        logger.info(
            "Please be sure the configuration file is correct: "
            "it exists and the format is correct (especially in "
            "the case of netrc or motuclient file)."
        )
        return None
    else:
        logger.info("No credentials found.")
        logger.info(
            "Please provide credentials as arguments or environment "
            "variables, or use the 'login' command to create a credentials file."
        )
    return None


def create_copernicusmarine_configuration_file(
    username: str,
    password: str,
    configuration_file_directory: pathlib.Path,
    force_overwrite: bool,
) -> tuple[Optional[pathlib.Path], bool]:
    configuration_lines = [
        "[credentials]\n",
        f"username={username}\n",
        f"password={password}\n",
    ]
    configuration_filename = pathlib.Path(
        configuration_file_directory / DEFAULT_CLIENT_CREDENTIALS_FILENAME
    )
    if configuration_filename.exists() and not force_overwrite:
        confirmed = click.confirm(
            f"File {configuration_filename} already exists, overwrite it ?"
        )
        if not confirmed:
            logger.error("Abort")
            return None, True

    configuration_file_directory.mkdir(parents=True, exist_ok=True)
    configuration_file = open(configuration_filename, "w")
    configuration_string = base64.b64encode(
        "".join(configuration_lines).encode("ascii", "strict")
    ).decode("utf8")
    configuration_file.write(configuration_string)
    configuration_file.close()
    return configuration_filename, False


def _check_credentials_with_old_cas(
    username: str, password: str
) -> Union[str, None]:
    logger.debug("Checking user credentials with old cas...")
    service = "copernicus-marine-client"
    cmems_cas_login_url = (
        f"{COPERNICUS_MARINE_MARINE_AUTH_OLD_SYSTEM_URL}?service={service}"
    )
    conn_session = get_configured_requests_session()
    logger.debug(f"GETing {cmems_cas_login_url}...")
    login_session = conn_session.get(
        cmems_cas_login_url, proxies=conn_session.proxies
    )
    login_session.raise_for_status()
    login_from_html = lxml.html.fromstring(login_session.text)
    hidden_elements_from_html = login_from_html.xpath(
        '//form//input[@type="hidden"]'
    )
    payload = {
        he.attrib["name"]: he.attrib["value"]
        for he in hidden_elements_from_html
    }
    payload["username"] = username
    payload["password"] = password
    logger.debug(f"POSTing credentials to {cmems_cas_login_url}...")
    login_response = conn_session.post(
        cmems_cas_login_url, data=payload, proxies=conn_session.proxies
    )
    login_response.raise_for_status()
    login_success = 'class="success"' in login_response.text
    logger.debug("User credentials checked")
    conn_session.close()
    return username if login_success else None


def _check_credentials_with_cas(
    username: str, password: str
) -> Union[str, None]:
    keycloak_url = COPERNICUS_MARINE_AUTH_SYSTEM_TOKEN_ENDPOINT
    client_id = "toolbox"
    scope = "openid profile email"

    data = {
        "client_id": client_id,
        "grant_type": "password",
        "username": username,
        "password": password,
        "scope": scope,
    }
    conn_session = get_configured_requests_session()
    logger.debug(f"POSTing credentials to {keycloak_url}...")
    response = conn_session.post(
        keycloak_url, data=data, proxies=conn_session.proxies
    )
    response.raise_for_status()
    if response.status_code == 200:
        token_response = response.json()
        access_token = token_response["access_token"]
        bearer_auth = BearerAuth(access_token)
        userinfo_url = COPERNICUS_MARINE_AUTH_SYSTEM_USERINFO_ENDPOINT
        logger.debug(f"GETing {userinfo_url}...")
        response = conn_session.get(
            userinfo_url, auth=bearer_auth, proxies=conn_session.proxies
        )
        response.raise_for_status()
        if response.status_code == 200:
            response_json = response.json()
            return response_json["preferred_username"]
    conn_session.close()
    return None


def _are_copernicus_marine_credentials_valid_old_system(
    username: str, password: str
) -> Union[str, None]:
    number_of_retry = 3
    while number_of_retry > 0:
        try:
            user = _check_credentials_with_old_cas(
                username=username, password=password
            )
            return user
        except requests.exceptions.ConnectTimeout:
            number_of_retry -= 1
        except requests.exceptions.ConnectionError:
            number_of_retry -= 1
    raise CouldNotConnectToAuthenticationSystem()


def _validate_and_get_user(username: str, password: str) -> Union[str, None]:
    try:
        result = _get_user_new_system(username, password)
        return result

    except Exception as e:
        logger.debug(
            f"Could not connect with new authentication system because of: {e}"
        )
        logger.debug("Trying with old authentication system...")
        return _are_copernicus_marine_credentials_valid_old_system(
            username, password
        )


def _get_user_new_system(username: str, password: str) -> Union[str, None]:
    number_of_retry = 3
    while number_of_retry > 0:
        try:
            user = _check_credentials_with_cas(
                username=username, password=password
            )
            return user
        except requests.exceptions.ConnectTimeout:
            number_of_retry -= 1
        except requests.exceptions.ConnectionError:
            number_of_retry -= 1
    raise CouldNotConnectToAuthenticationSystem()


def get_credential(
    credential: Optional[str],
    credential_type: Literal["username", "password"],
    hide_input: bool,
    credentials_file: Optional[pathlib.Path],
) -> str:
    if not credential:
        credential = _retrieve_credential_from_environment_variable(
            credential_type
        )
        if not credential:
            credential = _retrieve_credential_from_configuration_files(
                credential_type=credential_type,
                credentials_file=credentials_file,
            )
            if not credential:
                credential = _retrieve_credential_from_prompt(
                    credential_type, hide_input=hide_input
                )
                if not credential:
                    raise ValueError(f"{credential} cannot be None")
    else:
        logger.debug("Credentials loaded from function arguments")
    return credential


def get_and_check_username_password(
    username: Optional[str],
    password: Optional[str],
    credentials_file: Optional[pathlib.Path],
) -> tuple[str, str]:
    username, password = get_username_password(
        username=username, password=password, credentials_file=credentials_file
    )
    user = _validate_and_get_user(
        username,
        password,
    )
    if not user:
        raise InvalidUsernameOrPassword(
            "Learn how to recover your credentials at: "
            "https://help.marine.copernicus.eu/en/articles/"
            "4444552-i-forgot-my-username-or-my-password-what-should-i-do"
        )
    return (user, password)


def get_username_password(
    username: Optional[str],
    password: Optional[str],
    credentials_file: Optional[pathlib.Path],
) -> tuple[str, str]:
    username = get_credential(
        username,
        "username",
        hide_input=False,
        credentials_file=credentials_file,
    )
    password = get_credential(
        password,
        "password",
        hide_input=True,
        credentials_file=credentials_file,
    )
    return (username, password)


def _get_credential_from_environment_variable_or_prompt(
    credential: Optional[str],
    credential_type: Literal["username", "password"],
    hide_input: bool,
) -> str:
    if not credential:
        credential = _retrieve_credential_from_environment_variable(
            credential_type
        )
        if not credential:
            credential = _retrieve_credential_from_prompt(
                credential_type, hide_input
            )
            if not credential:
                raise CredentialsCannotBeNone(credential_type)
    return credential


def credentials_file_builder(
    username: Optional[str],
    password: Optional[str],
    configuration_file_directory: pathlib.Path,
    force_overwrite: bool,
) -> tuple[Optional[pathlib.Path], bool]:
    """
    Returns:
        a path to the configuration file: if none, the credentials are not valid

        a boolean that indicates
        if the user aborted the creation of the configuration file
    """
    username = _get_credential_from_environment_variable_or_prompt(
        username, "username", False
    )
    password = _get_credential_from_environment_variable_or_prompt(
        password, "password", True
    )
    copernicus_marine_credentials_are_valid = _validate_and_get_user(
        username, password
    )
    if copernicus_marine_credentials_are_valid:
        (
            configuration_file,
            has_been_aborted,
        ) = create_copernicusmarine_configuration_file(
            username=username,
            password=password,
            configuration_file_directory=configuration_file_directory,
            force_overwrite=force_overwrite,
        )
        if has_been_aborted:
            return configuration_file, has_been_aborted
        if configuration_file:
            return configuration_file, False
    return None, False



import pathlib

import click
from click import Context, Option, UsageError
from click.core import ParameterSource

from copernicusmarine.core_functions import documentation_utils
from copernicusmarine.core_functions.click_custom_class import (
    CustomDeprecatedClickOption,
)


class MutuallyExclusiveOption(Option):
    def __init__(self, *args, **kwargs):
        self.mutually_exclusive = set(kwargs.pop("mutually_exclusive", []))
        help = kwargs.get("help", "")
        if self.mutually_exclusive:
            ex_str = ", ".join(sorted(self.mutually_exclusive))
            kwargs["help"] = (
                help
                + """
                 NOTE: This argument is mutually exclusive with arguments: ["""
                + ex_str
                + "]."
            )

        super().__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        excluded_option_to_snake_case = {
            name.replace("-", "_") for name in self.mutually_exclusive
        }
        if (
            excluded_option_to_snake_case.intersection(opts)
            and self.name
            and self.name in opts
        ):
            raise UsageError(
                f"Illegal usage: arguments '{self.name.replace('_', '-')}' and "
                f"'{', '.join(self.mutually_exclusive)}'"
                " are mutually exclusive."
            )

        return super().handle_parse_result(ctx, opts, args)


class OtherOptionsPassedWithCreateTemplate(Exception):
    """
    Exception raised when other options are passed with create_template.

    Please note that create_template should be passed with no other option
    except log_level.
    """

    pass


def assert_cli_args_are_not_set_except_create_template(
    context: Context,
) -> None:
    for key in context.params:
        if key not in ["create_template", "log_level"]:
            parameter_source = context.get_parameter_source(key)
            if parameter_source != ParameterSource.DEFAULT:
                raise OtherOptionsPassedWithCreateTemplate(key)


tqdm_disable_option = click.option(
    "--disable-progress-bar",
    is_flag=True,
    default=False,
    help=documentation_utils.SHARED["DISABLE_PROGRESS_BAR_HELP"],
)

force_dataset_version_option = click.option(
    "--dataset-version",
    type=str,
    default=None,
    help=documentation_utils.SHARED["DATASET_VERSION_HELP"],
)

force_dataset_part_option = click.option(
    "--dataset-part",
    type=str,
    default=None,
    help=documentation_utils.SHARED["DATASET_PART_HELP"],
)

credentials_file_option = click.option(
    "--credentials-file",
    type=click.Path(path_type=pathlib.Path),
    help=documentation_utils.SHARED["CREDENTIALS_FILE_HELP"],
)

force_download_option = click.option(
    "--force-download",
    is_flag=True,
    default=False,
    hidden=True,
    cls=CustomDeprecatedClickOption,
    custom_deprecated=["--force-download"],
)



import importlib.util
import json
import logging
import pathlib
from datetime import datetime
from typing import List, Optional

from copernicusmarine.catalogue_parser.models import (
    CopernicusMarineServiceFormat,
    CopernicusMarineServiceNames,
)
from copernicusmarine.core_functions.credentials_utils import (
    get_and_check_username_password,
)
from copernicusmarine.core_functions.exceptions import ServiceNotSupported
from copernicusmarine.core_functions.marine_datastore_config import (
    get_config_and_check_version_subset,
)
from copernicusmarine.core_functions.models import (
    CommandType,
    CoordinatesSelectionMethod,
    ResponseSubset,
    VerticalAxis,
)
from copernicusmarine.core_functions.request_structure import (
    SubsetRequest,
    convert_motu_api_request_to_structure,
)
from copernicusmarine.core_functions.services_utils import (
    RetrievalService,
    get_retrieval_service,
)
from copernicusmarine.core_functions.utils import get_unique_filepath
from copernicusmarine.download_functions.download_sparse import download_sparse
from copernicusmarine.download_functions.download_zarr import download_zarr
from copernicusmarine.download_functions.subset_xarray import (
    check_dataset_subset_bounds,
)
from copernicusmarine.download_functions.utils import FileFormat
from copernicusmarine.versioner import __version__ as copernicusmarine_version

logger = logging.getLogger("copernicusmarine")


def subset_function(
    dataset_id: Optional[str],
    force_dataset_version: Optional[str],
    force_dataset_part: Optional[str],
    username: Optional[str],
    password: Optional[str],
    variables: Optional[List[str]],
    minimum_x: Optional[float],
    maximum_x: Optional[float],
    minimum_y: Optional[float],
    maximum_y: Optional[float],
    minimum_depth: Optional[float],
    maximum_depth: Optional[float],
    vertical_axis: VerticalAxis,
    start_datetime: Optional[datetime],
    end_datetime: Optional[datetime],
    platform_ids: Optional[List[str]],
    coordinates_selection_method: CoordinatesSelectionMethod,
    output_filename: Optional[str],
    file_format: FileFormat,
    force_service: Optional[str],
    request_file: Optional[pathlib.Path],
    output_directory: Optional[pathlib.Path],
    credentials_file: Optional[pathlib.Path],
    motu_api_request: Optional[str],
    overwrite: bool,
    skip_existing: bool,
    dry_run: bool,
    disable_progress_bar: bool,
    staging: bool,
    netcdf_compression_level: int,
    netcdf3_compatible: bool,
    chunk_size_limit: int,
    raise_if_updating: bool,
) -> ResponseSubset:
    marine_datastore_config = get_config_and_check_version_subset(staging)
    if staging:
        logger.warning(
            "Detecting staging flag for subset command. "
            "Data will come from the staging environment."
        )

    subset_request = SubsetRequest(dataset_id=dataset_id or "")
    if netcdf3_compatible:
        documentation_url = (
            f"https://toolbox-docs.marine.copernicus.eu"
            f"/en/v{copernicusmarine_version}/installation.html#dependencies"
        )
        assert importlib.util.find_spec("netCDF4"), (
            "To enable the NETCDF3_COMPATIBLE option, the 'netCDF4' "
            f"package is required. "
            f"Please see {documentation_url}."
        )
    if request_file:
        subset_request.from_file(request_file)
    if motu_api_request:
        motu_api_subset_request = convert_motu_api_request_to_structure(
            motu_api_request
        )
        subset_request.update(motu_api_subset_request.__dict__)
    request_update_dict = {
        "force_dataset_version": force_dataset_version,
        "force_dataset_part": force_dataset_part,
        "variables": variables,
        "minimum_x": minimum_x,
        "maximum_x": maximum_x,
        "minimum_y": minimum_y,
        "maximum_y": maximum_y,
        "minimum_depth": minimum_depth,
        "maximum_depth": maximum_depth,
        "vertical_axis": vertical_axis,
        "start_datetime": start_datetime,
        "end_datetime": end_datetime,
        "platform_ids": platform_ids,
        "coordinates_selection_method": coordinates_selection_method,
        "output_filename": output_filename,
        "file_format": file_format,
        "force_service": force_service,
        "output_directory": output_directory,
        "netcdf_compression_level": netcdf_compression_level,
        "netcdf3_compatible": netcdf3_compatible,
        "dry_run": dry_run,
        "raise_if_updating": raise_if_updating,
    }
    subset_request.update(request_update_dict)
    if not subset_request.dataset_id:
        raise ValueError("Please provide a dataset id for a subset request.")
    username, password = get_and_check_username_password(
        username,
        password,
        credentials_file,
    )
    # Specific treatment for default values:
    # In order to not overload arguments with default values
    if overwrite:
        subset_request.overwrite = overwrite
    if skip_existing:
        subset_request.skip_existing = skip_existing

    retrieval_service: RetrievalService = get_retrieval_service(
        subset_request.dataset_id,
        subset_request.force_dataset_version,
        subset_request.force_dataset_part,
        subset_request.force_service,
        CommandType.SUBSET,
        dataset_subset=subset_request,
        platform_ids_subset=bool(subset_request.platform_ids),
        marine_datastore_config=marine_datastore_config,
    )
    subset_request.dataset_url = retrieval_service.uri
    # TODO: Add check for insitu datasets
    check_dataset_subset_bounds(
        service=retrieval_service.service,
        part=retrieval_service.dataset_part,
        dataset_subset=subset_request,
        coordinates_selection_method=subset_request.coordinates_selection_method,
        axis_coordinate_id_mapping=retrieval_service.axis_coordinate_id_mapping,
    )
    if retrieval_service.service_name in [
        CopernicusMarineServiceNames.GEOSERIES,
        CopernicusMarineServiceNames.TIMESERIES,
        CopernicusMarineServiceNames.PLATFORMSERIES,
        CopernicusMarineServiceNames.OMI_ARCO,
        CopernicusMarineServiceNames.STATIC_ARCO,
    ]:
        if (
            retrieval_service.service_format
            == CopernicusMarineServiceFormat.ZARR
        ):
            raise_when_all_dataset_requested(subset_request, False)
            if subset_request.file_format not in ["netcdf", "zarr"]:
                raise ValueError(
                    f"{subset_request.file_format} is not a valid format "
                    "for this dataset. "
                    "Available format for this dataset is 'netcdf' or 'zarr'."
                )
            response = download_zarr(
                username,
                password,
                subset_request,
                retrieval_service.dataset_id,
                disable_progress_bar,
                retrieval_service.dataset_valid_start_date,
                retrieval_service.service,
                retrieval_service.is_original_grid,
                retrieval_service.axis_coordinate_id_mapping,
                chunk_size_limit,
                retrieval_service.dataset_chunking,
            )
        if (
            retrieval_service.service_format
            == CopernicusMarineServiceFormat.SQLITE
        ):
            raise_when_all_dataset_requested(subset_request, True)
            if subset_request.file_format not in ["parquet", "csv"]:
                logger.debug(
                    "Using 'csv' format by default. "
                    "'parquet' format can also be set with 'file-format' option."
                )
                subset_request.file_format = "csv"
            if subset_request.coordinates_selection_method not in [
                "inside",
                "strict-inside",
            ]:
                logger.warning(
                    f"coordinates-selection-method "
                    f"{subset_request.coordinates_selection_method} "
                    "is not supported for sparse data. "
                    "Using 'inside' by default."
                )
            response = download_sparse(
                username,
                subset_request,
                retrieval_service.metadata_url,
                retrieval_service.service,
                retrieval_service.axis_coordinate_id_mapping,
                retrieval_service.product_doi,
                disable_progress_bar,
            )
    else:
        raise ServiceNotSupported(retrieval_service.service_name)
    return response


def create_subset_template() -> None:
    filename = pathlib.Path("subset_template.json")
    if filename.exists():
        get_unique_filepath(
            filepath=filename,
        )
    with open(filename, "w") as output_file:
        json.dump(
            {
                "dataset_id": "cmems_mod_glo_phy_myint_0.083deg_P1M-m",
                "start_datetime": "2023-10-01",
                "end_datetime": "2023-11-01",
                "minimum_longitude": -85,
                "maximum_longitude": -10,
                "minimum_latitude": 35,
                "maximum_latitude": 43,
                "minimum_depth": 1,
                "maximum_depth": 10,
                "variables": ["so", "thetao"],
                "output_directory": "copernicusmarine_data",
                "service": False,
                "request_file": False,
                "motu_api_request": False,
                "overwrite": False,
                "dry_run": False,
            },
            output_file,
            indent=4,
        )
    logger.info(f"Template created at: {filename}")


def raise_when_all_dataset_requested(
    subset_request: SubsetRequest, sparse_data: bool
) -> None:
    """
    Raise an error if the subset request is not a subsetting request
    and all the dataset would be downloaded.

    Parameters
    ----------
    subset_request: SubsetRequest
        The subset request.
    sparse_data: bool
        If the requested dataset is sparse data.
        If yes, only subsetting on platform_ids is allowed.
        Otherwise it should raise.
    """
    if all(
        e is None
        for e in [
            subset_request.variables,
            subset_request.minimum_x,
            subset_request.maximum_x,
            subset_request.minimum_y,
            subset_request.maximum_y,
            subset_request.minimum_depth,
            subset_request.maximum_depth,
            subset_request.start_datetime,
            subset_request.end_datetime,
            (sparse_data and subset_request.platform_ids) or None,
        ]
    ):
        logger.info(
            "To retrieve a complete dataset, please use instead: "
            f"copernicusmarine get --dataset-id {subset_request.dataset_id}"
        )
        raise ValueError(
            "Missing subset option. Try 'copernicusmarine subset --help'."
        )



import pathlib
from typing import Optional

from copernicusmarine.core_functions.credentials_utils import (
    DEFAULT_CLIENT_BASE_DIRECTORY,
)
from copernicusmarine.core_functions.login import login_function


def login(
    username: Optional[str] = None,
    password: Optional[str] = None,
    configuration_file_directory: pathlib.Path = DEFAULT_CLIENT_BASE_DIRECTORY,
    force_overwrite: bool = False,
    check_credentials_valid: bool = False,
    credentials_file: Optional[pathlib.Path] = None,
) -> bool:
    """
    Create a configuration file with your Copernicus Marine credentials under the ``$HOME/.copernicusmarine`` directory (overwritable with the ``force_overwrite`` option).

    Parameters
    ----------
    username : str, optional
        If not set, search for environment variable COPERNICUSMARINE_SERVICE_USERNAME, else ask for user input.
    password : str, optional
        If not set, search for environment variable COPERNICUSMARINE_SERVICE_PASSWORD, else ask for user input.
    configuration_file_directory : Union[pathlib.Path, str]
        Path to the directory where the configuration file will be stored.
    force_overwrite : bool
        Flag to skip confirmation before overwriting configuration file.
    check_credentials_valid : bool
        Flag to check if the credentials are valid. No other action will be performed. The validity will be check in this order:
        1. Check if the credentials are valid with the provided username and password.
        2. Check if the credentials are valid in the environment variables.
        3. Check if the credentials are valid in the configuration file.
        When any is found (valid or not valid), will return immediately.
    credentials_file : Union[pathlib.Path, str], optional
        Path to a credentials file if not in its default directory (``$HOME/.copernicusmarine``). Accepts .copernicusmarine-credentials / .netrc or _netrc / motuclient-python.ini files. Will only be taken into account when checking the credentials validity.

    Returns
    -------
    bool
        True value if the login was successfully completed, False otherwise.
    """  # noqa
    return login_function(
        username=username,
        password=password,
        configuration_file_directory=configuration_file_directory,
        force_overwrite=force_overwrite,
        check_credentials_valid=check_credentials_valid,
        configuration_file=credentials_file,
    )



import logging
import time
from collections.abc import MutableMapping
from typing import Optional

import botocore.config
import botocore.exceptions
import botocore.session

from copernicusmarine.core_functions.sessions import (
    get_configured_boto3_session,
)

logger = logging.getLogger("copernicusmarine")


class CustomS3StoreZarrV2(MutableMapping):
    def __init__(
        self,
        endpoint: str,
        bucket: str,
        root_path: str,
        copernicus_marine_username: Optional[str] = None,
        number_of_retries: int = 9,
        initial_retry_wait_seconds: int = 1,
    ):
        self._root_path = root_path.lstrip("/")
        self._bucket = bucket
        self.client, _ = get_configured_boto3_session(
            endpoint,
            ["GetObject", "HeadObject", "ListObjectsV2"],
            copernicus_marine_username,
        )

        self.number_of_retries = number_of_retries
        self.initial_retry_wait_seconds = initial_retry_wait_seconds

    def __getitem__(self, key):
        def fn():
            full_key = f"{self._root_path}/{key}"
            try:
                resp = self.client.get_object(
                    Bucket=self._bucket, Key=full_key
                )
                return resp["Body"].read()
            except botocore.exceptions.ClientError as e:
                raise KeyError(key) from e

        return self.with_retries(fn)

    def __contains__(self, key):
        full_key = f"{self._root_path}/{key}"
        try:
            self.client.head_object(Bucket=self._bucket, Key=full_key)
            return True
        except botocore.exceptions.ClientError as e:
            if "404" in str(e) or "403" in str(e):
                return False
            raise

    def __setitem__(self, key, value, headers=None):
        def fn():
            full_key = f"{self._root_path}/{key}"
            final_headers = headers if headers is not None else {}
            self.client.put_object(
                Bucket=self._bucket, Key=full_key, Body=value, **final_headers
            )

        return self.with_retries(fn)

    def __delitem__(self, key):
        def fn():
            full_key = f"{self._root_path}/{key}"
            self.client.delete_object(Bucket=self._bucket, Key=full_key)

        return self.with_retries(fn)

    # Example of headers: {"ContentType": "application/json", "ContentEncoding": "gzip"}
    def set_item_with_headers(self, key, value, headers):
        return self.__setitem__(key, value, headers)

    def keys(self):
        keys = []
        cursor = self._root_path
        while True:
            resp = self.client.list_objects_v2(
                Bucket=self._bucket, Prefix=self._root_path, StartAfter=cursor
            )
            entries = resp.get("Contents", [])
            keys += [
                o["Key"].removeprefix(self._root_path).lstrip("/")
                for o in entries
            ]
            if not resp["IsTruncated"]:
                break
            cursor = entries[-1]["Key"]
        return keys

    def __iter__(self):
        keys = self.keys()
        yield from keys

    def __len__(self):
        return len(self.keys())

    def clear(self):
        keys = self.keys()
        idx = 0
        while idx < len(keys):
            some_keys = keys[idx : idx + 1000]
            objects = list(
                map(lambda k: {"Key": f"{self._root_path}/{k}"}, some_keys)
            )
            self.client.delete_objects(
                Bucket=self._bucket, Delete={"Objects": objects}
            )
            idx += 1000

    def with_retries(self, fn):
        retry_delay = self.initial_retry_wait_seconds
        for index_try in range(self.number_of_retries):
            try:
                return fn()
            # KeyError is a normal error that we want to propagate
            # (e.g. if we try to get a chunk and it doesn't exist,
            # we want the caller to know this has happened -- and not retry!)
            except KeyError:
                raise
            except Exception as e:
                if index_try == self.number_of_retries - 1:
                    raise e
                logger.debug(f"S3 error: {e}")
                logger.debug(f"Retrying in {retry_delay} s...")
                time.sleep(retry_delay)
                retry_delay *= 2



import logging
import time
from typing import AsyncIterator, Iterable, Optional, Union

import botocore.config
import botocore.exceptions
import botocore.session
from zarr.abc.store import ByteRequest, Store
from zarr.core.buffer import Buffer, BufferPrototype, default_buffer_prototype
from zarr.core.common import BytesLike

from copernicusmarine.core_functions.sessions import (
    get_configured_boto3_session,
)

logger = logging.getLogger("copernicusmarine")


class CustomS3StoreZarrV3(Store):
    def __init__(
        self,
        endpoint: str,
        bucket: str,
        root_path: str,
        copernicus_marine_username: Optional[str] = None,
        number_of_retries: int = 9,
        initial_retry_wait_seconds: int = 1,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._root_path = root_path.lstrip("/")
        self._bucket = bucket
        self.client, _ = get_configured_boto3_session(
            endpoint,
            ["GetObject", "HeadObject", "ListObjectsV2"],
            copernicus_marine_username,
        )

        self.number_of_retries = number_of_retries
        self.initial_retry_wait_seconds = initial_retry_wait_seconds

    def __eq__(self, value: object) -> bool:
        """Equality comparison."""
        return (
            isinstance(value, CustomS3StoreZarrV3)
            and self._root_path == value._root_path
            and self._bucket == value._bucket
            and self.client.meta.endpoint_url == value.client.meta.endpoint_url
        )

    async def get(
        self,
        key: str,
        prototype: BufferPrototype,
        byte_range: Union[ByteRequest, None] = None,
    ) -> Union[Buffer, None]:
        def fn():
            full_key = f"{self._root_path}/{key}"
            try:
                resp = self.client.get_object(
                    Bucket=self._bucket, Key=full_key
                )
                res = resp["Body"].read()
                return prototype.buffer.from_bytes(res)
            except botocore.exceptions.ClientError as e:
                https_status_code: int = e.response["ResponseMetadata"][
                    "HTTPStatusCode"
                ]
                if https_status_code == 404 or https_status_code == 403:
                    # means the object does not exist
                    return None
                else:
                    raise e

        return self.with_retries(fn)

    async def get_partial_values(
        self,
        prototype: BufferPrototype,
        key_ranges: Iterable[tuple[str, Union[ByteRequest, None]]],
    ) -> list[Union[Buffer, None]]:
        return [
            await self.get(key, prototype, byte_range)
            for key, byte_range in key_ranges
        ]

    async def exists(self, key: str) -> bool:
        return self.get(key, default_buffer_prototype()) is not None

    def supports_writes(self) -> bool:
        return False

    async def set(self, key: str, value: Buffer) -> None:
        pass

    def supports_deletes(self) -> bool:
        return False

    async def delete(self, key: str) -> None:
        pass

    def supports_partial_writes(self) -> bool:
        return False

    async def set_partial_values(
        self, key_start_values: Iterable[tuple[str, int, BytesLike]]
    ) -> None:
        pass

    def supports_listing(self) -> bool:
        """Does the store support listing?"""
        return True

    def keys(self, prefix: str = "") -> list[str]:
        keys = []
        cursor = self._root_path
        while True:
            resp = self.client.list_objects_v2(
                Bucket=self._bucket,
                Prefix=self._root_path + prefix,
                StartAfter=cursor,
            )
            entries = resp.get("Contents", [])
            keys += [
                o["Key"].removeprefix(self._root_path).lstrip("/")
                for o in entries
            ]
            if not resp["IsTruncated"]:
                break
            cursor = entries[-1]["Key"]
        return keys

    async def list(self) -> AsyncIterator[str]:
        for key in self.keys():
            yield key

    async def list_prefix(self, prefix: str) -> AsyncIterator[str]:
        for key in self.keys(prefix):
            yield key

    async def list_dir(self, prefix: str) -> AsyncIterator[str]:
        for key in self.keys(prefix):
            yield key

    def with_retries(self, fn):
        retry_delay = self.initial_retry_wait_seconds
        for index_try in range(self.number_of_retries):
            try:
                return fn()
            except Exception as e:
                if index_try == self.number_of_retries - 1:
                    raise e
                logger.debug(f"S3 error: {e}")
                logger.debug(f"Retrying in {retry_delay} s...")
                time.sleep(retry_delay)
                retry_delay *= 2



import logging
import pathlib
from typing import Optional

import click

from copernicusmarine.command_line_interface.exception_handler import (
    log_exception_and_exit,
)
from copernicusmarine.core_functions import documentation_utils
from copernicusmarine.core_functions.click_custom_class import (
    CustomClickOptionsCommand,
)
from copernicusmarine.core_functions.credentials_utils import (
    DEFAULT_CLIENT_BASE_DIRECTORY,
)
from copernicusmarine.core_functions.login import login_function

logger = logging.getLogger("copernicusmarine")


@click.group()
def cli_login() -> None:
    pass


@cli_login.command(
    "login",
    cls=CustomClickOptionsCommand,
    short_help="Create a configuration file with your Copernicus Marine credentials.",
    help=documentation_utils.LOGIN["LOGIN_DESCRIPTION_HELP"]
    + " \n\nReturns\n "
    + documentation_utils.LOGIN["LOGIN_RESPONSE_HELP"],  # noqa
    epilog="""
    Examples:

    Using environment variables:

    .. code-block:: bash

        COPERNICUSMARINE_SERVICE_USERNAME=<USERNAME> COPERNICUSMARINE_SERVICE_PASSWORD=<PASSWORD> copernicusmarine login

    Using command line arguments:

    .. code-block:: bash

        copernicusmarine login --username <USERNAME> --password <PASSWORD>

    Using directly user input:

    .. code-block:: bash

        copernicusmarine login
        > Username: [USER-INPUT]
        > Password: [USER-INPUT]
    """,  # noqa
)
@click.option(
    "--username",
    hide_input=False,
    help=documentation_utils.LOGIN["USERNAME_HELP"],
)
@click.option(
    "--password",
    hide_input=True,
    help=documentation_utils.LOGIN["PASSWORD_HELP"],
)
@click.option(
    "--configuration-file-directory",
    type=click.Path(path_type=pathlib.Path),
    default=DEFAULT_CLIENT_BASE_DIRECTORY,
    help=documentation_utils.LOGIN["CONFIGURATION_FILE_DIRECTORY_HELP"],
)
@click.option(
    "--force-overwrite",
    is_flag=True,
    default=False,
    help=documentation_utils.LOGIN["FORCE_OVERWRITE_HELP"],
)
@click.option(
    "--check-credentials-valid",
    is_flag=True,
    default=False,
    help=documentation_utils.LOGIN["CHECK_CREDENTIALS_VALID_HELP"],
)
@click.option(
    "--credentials-file",
    type=click.Path(path_type=pathlib.Path),
    help=documentation_utils.LOGIN["CREDENTIALS_FILE_HELP"],
)
@click.option(
    "--log-level",
    type=click.Choice(["DEBUG", "INFO", "WARN", "ERROR", "CRITICAL", "QUIET"]),
    default="INFO",
    help=documentation_utils.LOGIN["LOG_LEVEL_HELP"],
)
@log_exception_and_exit
def login(
    username: Optional[str],
    password: Optional[str],
    configuration_file_directory: pathlib.Path,
    force_overwrite: bool,
    check_credentials_valid: bool,
    credentials_file: Optional[pathlib.Path],
    log_level: str = "INFO",
) -> None:
    if log_level == "QUIET":
        logger.disabled = True
        logger.setLevel(level="CRITICAL")
    else:
        logger.setLevel(level=log_level)
    if not login_function(
        username=username,
        password=password,
        configuration_file_directory=configuration_file_directory,
        force_overwrite=force_overwrite,
        check_credentials_valid=check_credentials_valid,
        configuration_file=credentials_file,
    ):
        exit(1)



import logging
import pathlib
import shutil
import warnings
from collections import defaultdict
from copy import deepcopy
from typing import Optional

import pandas as pd
from arcosparse import (
    Entity,
    UserConfiguration,
    get_entities,
    subset_and_return_dataframe,
)

from copernicusmarine.catalogue_parser.models import CopernicusMarineService
from copernicusmarine.core_functions.environment_variables import (
    COPERNICUSMARINE_DISABLE_SSL_CONTEXT,
    COPERNICUSMARINE_SET_SSL_CERTIFICATE_PATH,
)
from copernicusmarine.core_functions.exceptions import (
    NotEnoughPlatformMetadata,
    WrongPlatformID,
)
from copernicusmarine.core_functions.models import (  # TimeExtent,
    FileStatus,
    ResponseSubset,
    StatusCode,
    StatusMessage,
    VerticalAxis,
)
from copernicusmarine.core_functions.request_structure import SubsetRequest
from copernicusmarine.core_functions.sessions import TRUST_ENV
from copernicusmarine.core_functions.utils import (
    construct_query_params_for_marine_data_store_monitoring,
    datetime_to_isoformat,
    get_unique_filepath,
    timestamp_parser,
)
from copernicusmarine.download_functions.utils import (
    build_filename_from_request,
    get_file_extension,
)

logger = logging.getLogger("copernicusmarine")

COLUMNS_RENAME = {
    "entity_id": "platform_id",
    "entity_type": "platform_type",
}

COLUMNS_ORDER_DEPTH = [
    "variable",
    "platform_id",
    "platform_type",
    "time",
    "longitude",
    "latitude",
    "depth",
    "pressure",
    "is_depth_from_producer",
    "value",
    "value_qc",
    "institution",
    "doi",
    "product_doi",
]

COLUMNS_ORDER_ELEVATION = deepcopy(COLUMNS_ORDER_DEPTH)
COLUMNS_ORDER_ELEVATION[COLUMNS_ORDER_ELEVATION.index("depth")] = "elevation"

SORTING = {
    "variable": True,
    "platform_id": True,
    "platform_type": True,
    "time": True,
}


# TODO: should we support necdf?
# https://stackoverflow.com/questions/46476920/xarray-writing-to-netcdf-from-pandas-dimension-issue # noqa
def download_sparse(
    username: str,
    subset_request: SubsetRequest,
    metadata_url: str,
    service: CopernicusMarineService,
    axis_coordinate_id_mapping: dict[str, str],
    product_doi: Optional[str],
    disable_progress_bar: bool,
) -> ResponseSubset:

    if subset_request.dry_run:
        _, variables, platform_ids = _read_dataframe_sparse(
            username,
            subset_request,
            metadata_url,
            service,
            product_doi,
            disable_progress_bar,
            dry_run=True,
        )
        response = _get_response_subset(
            subset_request,
            variables,
            platform_ids,
            axis_coordinate_id_mapping,
        )
        response.status = StatusCode.DRY_RUN
        response.message = StatusMessage.DRY_RUN
        return response

    df, variables, platform_ids = _read_dataframe_sparse(
        username,
        subset_request,
        metadata_url,
        service,
        product_doi,
        disable_progress_bar,
    )
    response = _get_response_subset(
        subset_request,
        variables,
        platform_ids,
        axis_coordinate_id_mapping,
    )
    output_path = response.file_path
    if subset_request.skip_existing and output_path.exists():
        response.file_status = FileStatus.IGNORED
        return response
    elif (
        subset_request.overwrite
        and output_path.exists()
        and output_path.is_dir()
    ):
        shutil.rmtree(output_path)

    if subset_request.output_directory:
        subset_request.output_directory.mkdir(parents=True, exist_ok=True)
    if subset_request.file_format == "parquet":
        df.to_parquet(output_path, index=False)
    else:
        df.to_csv(output_path, index=False)

    return response


def read_dataframe_sparse(
    username: str,
    subset_request: SubsetRequest,
    metadata_url: str,
    service: CopernicusMarineService,
    product_doi: Optional[str],
    disable_progress_bar: bool,
) -> pd.DataFrame:
    df, _, _ = _read_dataframe_sparse(
        username,
        subset_request,
        metadata_url,
        service,
        product_doi,
        disable_progress_bar,
    )
    return df


def _read_dataframe_sparse(
    username: str,
    subset_request: SubsetRequest,
    metadata_url: str,
    service: CopernicusMarineService,
    product_doi: Optional[str],
    disable_progress_bar: bool,
    dry_run: bool = False,
) -> tuple[pd.DataFrame, list[str], list[str]]:
    """
    Shared function for subset and read dataframe

    Returns also the variables and the platform_ids
    """
    user_configuration = _get_user_configuration(username)
    platforms_metadata = {
        entity.entity_id: entity
        for entity in get_entities(metadata_url, user_configuration)
    }
    if subset_request.platform_ids:
        platform_ids = _get_plaform_ids_to_subset(
            subset_request.platform_ids or [],
            set(platforms_metadata.keys()),
            service,
        )
    else:
        platform_ids = []
    variables = subset_request.variables or [
        variable.short_name for variable in service.variables
    ]
    if dry_run:
        return pd.DataFrame(), variables, platform_ids
    # see https://github.com/pandas-dev/pandas/issues/55928
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=FutureWarning)
        df = subset_and_return_dataframe(
            minimum_latitude=subset_request.minimum_y,
            maximum_latitude=subset_request.maximum_y,
            minimum_longitude=subset_request.minimum_x,
            maximum_longitude=subset_request.maximum_x,
            minimum_elevation=(
                -subset_request.maximum_depth
                if subset_request.maximum_depth is not None
                else None
            ),
            maximum_elevation=(
                -subset_request.minimum_depth
                if subset_request.minimum_depth is not None
                else None
            ),
            minimum_time=(
                subset_request.start_datetime.timestamp()
                if subset_request.start_datetime is not None
                else None
            ),
            maximum_time=(
                subset_request.end_datetime.timestamp()
                if subset_request.end_datetime is not None
                else None
            ),
            variables=variables,
            entities=platform_ids,
            vertical_axis=subset_request.vertical_axis,
            url_metadata=metadata_url,
            user_configuration=user_configuration,
            disable_progress_bar=disable_progress_bar,
            columns_rename=COLUMNS_RENAME,
        )

        df = _transform_dataframe(
            df,
            subset_request.vertical_axis,
            platforms_metadata,
            product_doi,
        )
    if df.empty:
        logger.info(
            "No data found for the given parameters. "
            "Please check your request and try again."
        )
    return (
        df,
        variables,
        platform_ids,
    )


def _get_user_configuration(username: str) -> UserConfiguration:
    return UserConfiguration(
        disable_ssl=COPERNICUSMARINE_DISABLE_SSL_CONTEXT == "True",
        trust_env=TRUST_ENV,
        ssl_certificate_path=COPERNICUSMARINE_SET_SSL_CERTIFICATE_PATH,
        max_concurrent_requests=20,
        extra_params=construct_query_params_for_marine_data_store_monitoring(
            username
        ),
    )


def _get_plaform_ids_to_subset(
    platform_ids: list[str],
    platforms_metadata_names: set[str],
    retrieval_service: CopernicusMarineService,
) -> list[str]:
    platforms_to_subset = []
    if not platforms_metadata_names:
        raise NotEnoughPlatformMetadata()
    platforms_names_with_types: set[str] = set()
    platforms_without_types_mapping: dict[str, list] = defaultdict(list)
    for platform_name in platforms_metadata_names:
        platform_name_without_type = platform_name.split("___")[0]
        platforms_without_types_mapping[platform_name_without_type].append(
            platform_name
        )
        platforms_names_with_types.add(platform_name)
    for platform_id in platform_ids:
        if platform_id in platforms_names_with_types:
            platforms_to_subset.append(platform_id)
        if platform_id in platforms_without_types_mapping:
            platforms_to_subset.extend(
                platforms_without_types_mapping[platform_id]
            )
    if not platforms_to_subset:
        raise WrongPlatformID(
            platform_ids, retrieval_service.platforms_metadata
        )
    return platforms_to_subset


def _build_filename_and_output_path(
    subset_request: SubsetRequest,
    variables: list[str],
    platform_ids: list[str],
    axis_coordinate_id_mapping: dict[str, str],
) -> tuple[pathlib.Path, pathlib.Path]:
    extension_file = get_file_extension(subset_request.file_format)
    filename = pathlib.Path(
        subset_request.output_filename
        or build_filename_from_request(
            subset_request,
            variables,
            platform_ids,
            axis_coordinate_id_mapping,
        )
    )

    if filename.suffix != extension_file:
        filename = pathlib.Path(f"{filename}{extension_file}")
    output_path = pathlib.Path(
        subset_request.output_directory,
        filename,
    )
    if not subset_request.overwrite and not subset_request.skip_existing:
        output_path = get_unique_filepath(output_path)
    return filename, output_path


def _get_response_subset(
    subset_request: SubsetRequest,
    variables: list[str],
    platform_ids: list[str],
    axis_coordinate_id_mapping: dict[str, str],
) -> ResponseSubset:
    filename, output_path = _build_filename_and_output_path(
        subset_request,
        variables,
        platform_ids,
        axis_coordinate_id_mapping,
    )
    return ResponseSubset(
        file_path=output_path,
        output_directory=subset_request.output_directory,
        filename=str(filename),
        file_size=None,
        data_transfer_size=None,
        variables=variables,
        # TODO: handle thoses extents maybe opening the dataframe
        coordinates_extent=[],
        status=StatusCode.SUCCESS,
        message=StatusMessage.SUCCESS,
        file_status=FileStatus.DOWNLOADED,
    )


def _transform_dataframe(
    df: pd.DataFrame,
    vertical_axis: VerticalAxis,
    platforms_metadata: dict[str, Entity],
    product_doi: Optional[str],
) -> pd.DataFrame:
    """
    Transform the dataframe to match the expected format to be consistent with MyOceanPro
    and Copernicus Marine Services.
    """  # noqa
    if df.empty:
        return df
    # Needs to be done before striping the type to the platform_id
    df["institution"] = df["platform_id"].apply(
        lambda x: (
            (
                platforms_metadata[x].institution
                if x in platforms_metadata
                else pd.NA
            )
            or pd.NA
        )
    )
    df["doi"] = df["platform_id"].apply(
        lambda x: (
            (platforms_metadata[x].doi if x in platforms_metadata else pd.NA)
            or pd.NA
        )
    )

    df["product_doi"] = (
        f"https://doi.org/{product_doi}"
        if product_doi and "https://" not in product_doi
        else product_doi
    )

    # From "platform___type" to "platform" since the type is in
    # column platform_type
    df["platform_id"] = df["platform_id"].str.split("___").str[0]

    df["time"] = df["time"].apply(
        lambda x: datetime_to_isoformat(timestamp_parser(x, unit="s"))
    )

    # Some depth values comes from the arcoification of the data
    # and are calculated from the pressure some others come
    # directly from the producer (ie native/original data)
    df["is_depth_from_producer"] = df["is_approx_elevation"].apply(
        lambda x: 0 if x else 1
    )
    df.drop(columns=["is_approx_elevation"], inplace=True)

    if vertical_axis == "elevation":
        df = df[COLUMNS_ORDER_ELEVATION]
    else:
        df = df[COLUMNS_ORDER_DEPTH]

    df.sort_values(
        by=list(SORTING.keys()),
        ascending=list(SORTING.values()),
        inplace=True,
    )

    df.reset_index(drop=True, inplace=True)
    return df



from functools import wraps
from typing import Any, Callable, TypeVar

import click

T = TypeVar("T", bound=Callable[..., Any])


def log_exception_and_exit(function: T) -> T:
    @wraps(function)
    def wrapper(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except click.Abort:
            print("Abort")
        except Exception as exception:
            raise exception

    return wrapper  # type: ignore



import logging
from functools import wraps
from itertools import chain
from re import finditer
from sys import exit
from typing import Callable

logger = logging.getLogger("copernicusmarine")


def _log_exception(log_function: Callable, exception: Exception):
    exception_string = str(exception).strip('"')
    details = f": {exception_string}" if exception_string else ""
    message = _exception_to_sentence(exception) + details
    log_function(message)


def log_exception_debug(exception: Exception):
    _log_exception(logger.debug, exception)


def log_exception_and_exit(function: Callable):
    @wraps(function)
    def wrapper(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except Exception as exception:
            try:
                custom_exception_message = exception.__getattribute__(
                    "custom_exception_message"
                )
                logger.error(custom_exception_message)
            except AttributeError:
                _log_exception(logger.error, exception)
            exit(1)

    return wrapper


def _exception_to_sentence(exception: Exception) -> str:
    return _camel_case_to_sentence(exception.__class__.__name__)


def _title_case_to_lower_case(identifier: str) -> str:
    return identifier.lower() if identifier.istitle() else identifier


def _camel_case_to_sentence(identifier: str) -> str:
    matches = finditer(
        ".+?(?:(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])|$)", identifier
    )
    capital_words = list(map(lambda match: match.group(0), matches))
    first_word = capital_words[:1]
    other_words = capital_words[1:]
    sentence = chain(first_word, map(_title_case_to_lower_case, other_words))
    return " ".join(sentence)



import logging
from typing import (
    Literal,
    Optional,
    Type,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

from pydantic import BaseModel

logger = logging.getLogger("copernicusmarine")


def check_type_is_base_model(type_to_check: Type) -> bool:
    try:
        return issubclass(type_to_check, BaseModel)
    except TypeError:
        return False


class WrongFieldsError(Exception):
    """
    Exception raised when the fields requested are not queryable.

    Please verify the fields requested are a field name in the response type model
    for the command you are using. See the response type page of `the documentation <https://toolbox-docs.marine.copernicus.eu/>`_ for more information.
    """  # noqa: E501

    def __init__(
        self,
        wrong_fields: set[str],
        available_fields: set[str],
        option_name: str,
        *args: object,
    ) -> None:
        super().__init__(*args)
        self.wrong_fields = wrong_fields
        self.available_fields = available_fields
        self.option_name = option_name

    def __str__(self) -> str:
        return (
            f"All ``{self.option_name}`` fields are invalid"
            f": {', '.join(sorted(self.wrong_fields))}.\n"
            f"Available fields: {', '.join(sorted(self.available_fields))}"
        )


def build_query(
    fields_to_include_or_exclude: set[str],
    type_to_check: Type,
    query: Optional[dict] = None,
) -> dict:
    """
    Recursively builds a query to include or exclude fields from a Pydantic model
    Allows to handle nested models, lists and dictionaries.

    Example of resulting query:
    input:
    build_query({"product_id", "dataset_id"}, CopernicusMarineCatalogue)

    output:
    {
        "products": {
            "__all__": {
                "product_id": True
                "datasets": {
                    "__all__": {
                        "dataset_id": True
                    }
                }
            },
        }
    }
    """  # noqa: E501
    if query is None:
        query = {}

    for (
        field_name,
        field_type,
    ) in get_type_hints(type_to_check).items():
        if field_name in fields_to_include_or_exclude:
            query[field_name] = True
            continue
        all_base_models = _get_base_models_in_type(field_type)
        for base_model, in_an_iterable in (all_base_models or {}).items():
            if field_name not in query:
                query[field_name] = {"__all__": {}} if in_an_iterable else {}
            result = build_query(
                fields_to_include_or_exclude,
                base_model,
                (
                    query[field_name]["__all__"]
                    if in_an_iterable
                    else query[field_name]
                ),
            )
            if not result:
                del query[field_name]
    return query


def _get_base_models_in_type(
    type_to_check: Type,
    in_an_iterable: bool = False,
) -> Optional[dict[Type, Literal["__all__", None]]]:
    models = {}
    if check_type_is_base_model(type_to_check):
        return {type_to_check: "__all__" if in_an_iterable else None}
    elif get_origin(type_to_check) is list:
        result = _get_base_models_in_type(
            get_args(type_to_check)[0], in_an_iterable=True
        )
        if result:
            models.update(result)
    elif get_origin(type_to_check) is dict:
        result = _get_base_models_in_type(
            get_args(type_to_check)[1], in_an_iterable=True
        )
        if result:
            models.update(result)
    elif get_origin(type_to_check) is Union:
        for union_type in get_args(type_to_check):
            result = _get_base_models_in_type(
                union_type, in_an_iterable=in_an_iterable
            )
            if result:
                models.update(result)
    return models


def _return_available_fields(
    type_to_check: Type, available_fields: Optional[set[str]] = None
) -> set[str]:
    """
    Recursively get all the fields that are available in the model
    """
    if available_fields is None:
        available_fields = set()

    for (
        field_name,
        field_type,
    ) in get_type_hints(type_to_check).items():
        available_fields.add(field_name)
        all_base_models = _get_base_models_in_type(field_type)
        for base_model, _ in (all_base_models or {}).items():
            result = _return_available_fields(base_model, available_fields)
            available_fields.union(result)
    return available_fields


def get_queryable_requested_fields(
    requested_fields: set[str], type_to_check: Type, option_name: str
) -> set[str]:
    """
    Computes all the fields that are available in the model and checks if the requested fields are queryable

    if none of the fields are queryable, it raises an error
    if some of the fields are not queryable, it returns the queryable fields and logs a warning
    Returns:
    - a set with the queryable fields
    """  # noqa: E501
    available_fields = _return_available_fields(type_to_check)
    queryable_fields = requested_fields.intersection(available_fields)
    wrong_inputs = requested_fields - queryable_fields
    if queryable_fields == set():
        raise WrongFieldsError(wrong_inputs, available_fields, option_name)
    elif wrong_inputs:
        logger.warning(
            f"Some ``{option_name}`` fields are invalid:"
            f" {', '.join(wrong_inputs)}.\n "
            f"Available fields: {', '.join(sorted(available_fields))}."
        )
    return queryable_fields
